Neko 1.99.6
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
fusedcg_aux.cu
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2025, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include "fusedcg_kernel.h"
37#include <device/cuda/check.h>
38#include <device/cuda/buffer.h>
39
40#ifdef HAVE_NVSHMEM
41#include <nvshmem.h>
42#include <nvshmemx.h>
43#endif
44
50
51extern "C" {
52
55
56#ifdef HAVE_NCCL
59#endif
60
61 void cuda_fusedcg_update_p(void *p, void *z, void *po, real *beta, int *n) {
62
63 const dim3 nthrds(1024, 1, 1);
64 const dim3 nblcks(((*n)+1024 - 1)/ 1024, 1, 1);
65 const cudaStream_t stream = (cudaStream_t) glb_cmd_queue;
66
68 <<<nblcks, nthrds, 0, stream>>>((real *) p, (real *) z,
69 (real *) po, *beta, *n);
71
72 }
73
74 void cuda_fusedcg_update_x(void *x, void *p, void *alpha, int *p_cur, int *n) {
75
76 const dim3 nthrds(1024, 1, 1);
77 const dim3 nblcks(((*n)+1024 - 1)/ 1024, 1, 1);
78 const cudaStream_t stream = (cudaStream_t) glb_cmd_queue;
79
81 <<<nblcks, nthrds, 0, stream>>>((real *) x, (const real **) p,
82 (const real *) alpha, *p_cur, *n);
84 }
85
86
87 real cuda_fusedcg_part2(void *a, void *b, void *c,
88 void *alpha_d , real *alpha, int *p_cur, int * n) {
89
90 const dim3 nthrds(1024, 1, 1);
91 const dim3 nblcks(((*n)+1024 - 1)/ 1024, 1, 1);
92 const int nb = ((*n) + 1024 - 1)/ 1024;
93 const cudaStream_t stream = (cudaStream_t) glb_cmd_queue;
94
95 /* nb + 1 to fit the two-element host staging area for any nb */
99
100 /* Store alpha(p_cur) in pinned memory */
101 fusedcg_buf[1] = (*alpha);
102
103 /* Update alpha_d(p_cur) = alpha(p_cur) */
104 real *alpha_d_p_cur = ((real *) alpha_d) + ((*p_cur - 1));
106 cudaMemcpyHostToDevice, stream));
107
108
110 <<<nblcks, nthrds, 0, stream>>>((real *) a, (real *) b,
111 (real *) c, *alpha,
112 (real *) fusedcg_buf_d, * n);
114
115 reduce_kernel<real><<<1, 1024, 0, stream>>>((real *) fusedcg_buf_d, nb);
117
118#ifdef HAVE_NCCL
120 DEVICE_NCCL_SUM, stream);
122 cudaMemcpyDeviceToHost, stream));
123 cudaStreamSynchronize(stream);
124#elif HAVE_NVSHMEM
125 if (sizeof(real) == sizeof(float)) {
127 (float *) fusedcg_buf_d,
128 (float *) fusedcg_buf_d, 1, stream);
129 }
130 else if (sizeof(real) == sizeof(double)) {
132 (double *) fusedcg_buf_d,
133 (double *) fusedcg_buf_d, 1, stream);
134
135 }
137 sizeof(real), cudaMemcpyDeviceToHost, stream));
138 cudaStreamSynchronize(stream);
139#elif HAVE_DEVICE_MPI
140 cudaStreamSynchronize(stream);
142 sizeof(real), DEVICE_MPI_SUM);
143#else
145 cudaMemcpyDeviceToHost, stream));
146 cudaStreamSynchronize(stream);
147#endif
148
149 return fusedcg_buf[0];
150 }
151}
void cuda_buffer_reserve(cuda_buffer_t *buf, size_t size)
Definition buffer.cu:50
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
#define CUDA_BUFFER_INIT_SYMM
Definition buffer.h:63
__global__ void const T *__restrict__ x
#define CUDA_CHECK(err)
Definition check.h:6
double real
#define DEVICE_MPI_SUM
void device_mpi_allreduce(void *buf_d, void *buf, int count, int nbytes, int op)
#define DEVICE_NCCL_SUM
void device_nccl_allreduce(void *sbuf_d, void *rbuf_d, int count, int nbytes, int op, void *stream)
real cuda_fusedcg_part2(void *a, void *b, void *c, void *alpha_d, real *alpha, int *p_cur, int *n)
void cuda_fusedcg_update_x(void *x, void *p, void *alpha, int *p_cur, int *n)
cuda_buffer_t fusedcg_redbuf
void cuda_fusedcg_update_p(void *p, void *z, void *po, real *beta, int *n)
void * host
Definition buffer.h:52
void * dev
Definition buffer.h:53